/*
 *  REGEXP, {akii-root}/src/regexp.c
 *
 * ==========================================================================
 *
 *                      Arena Kliens II. Releases
 *
 * ==========================================================================
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A legkiralyabb szovegszerkeszto: VIM - Vi IMproved 5  <http://www.vim.org>
 */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "debug.h"

typedef struct rexelem {
    int db;
    char **s;
} REXELEM;

char *rstrchr(char *miben, char *mit, int *h)
{
    *h = 0;
    while (*miben) {
	if (*miben != '\\') {
	    switch (*miben) {
		case '`':
		case '~':
		case '*':
		case '?':
		case '%':
		case '.':
		case '[':
		    *mit = *miben;
		    return (miben);
	    }
	    miben++;
	}
	else
	    miben += 2;
	(*h)++;
    }
    return (NULL);
}
int rstrncmp(char *mit, char *mivel, int h, int kisnagy)
{
    int i;

    for (i = 0; i < h; i++, *mit++, *mivel++) {
	if (*mivel == '\\')
	    mivel++;
	if ((!*mivel) || (!*mit))
	    return (*mit - *mivel);
	if (kisnagy) {
	    if (*mivel != *mit)
		return (1);
	}
	else if (toupper(*mivel) != toupper(*mit))
	    return (1);
    }
    return (0);
}
char *rstrstr(char *miben, char *mit, int kisnagy)
{
    int ok = 0, i = 0, j, lmiben, lmit;

    if (kisnagy)
	return (strstr(miben, mit));
    lmiben = strlen(miben);
    lmit = strlen(mit);
    while ((i <= lmiben - lmit) && (!ok)) {
	j = 0;
	while ((j < lmit) && (toupper(mit[j]) == toupper(miben[i + j])))
	    j++;
	ok = j >= lmit;
	if (!ok)
	    i++;
    }
    if (ok)
	return (miben + i);
    else
	return (NULL);
}
void tedd(char *innen, int ennyit, REXELEM *idem, int dik)
{
    int i;

    if (!dik)
	idem->s = malloc(sizeof(char *));

    else
	idem->s = realloc(idem->s, (dik + 1) * sizeof(char *));

    idem->s[dik] = malloc(ennyit + 1);
    for (i = 0; i < ennyit; i++)
	idem->s[dik][i] = innen[i];
    idem->s[dik][i] = 0;
    idem->db = dik + 1;
    return;
}
void torol(REXELEM *m)
{
    int i;

    for (i = 0; i < m->db; i++)
	free(m->s[i]);
    free(m->s);
    m->db = 0;
    m->s = NULL;
    return;
}
int regszavak(char *s, REXELEM *szavak)
{
    char *p;
    int i = 0;

#ifdef AKII_DEBUG
    dbgenter(dbg_regszavak, s);
#endif
    if (!szavak) {
#ifdef AKII_DEBUG
	dbgleave(dbg_regszavak);
#endif
	return (-1);
    }
    torol(szavak);
    do {
	p = s;
	while (isspace(*p))
	    p++;
	if (!(*p))
	    break;
	s = p;
	while ((!isspace(*s)) && (*s))
	    s++;
	tedd(p, s - p, szavak, i);
	i++;
    } while (*s);
#ifdef AKII_DEBUG
    dbgleave(dbg_regszavak);
#endif
    return (i);
}
int regqszavak(char *sz, REXELEM *qszavak)
{
    char w[1024], *s = w + 1, *p, *q = w;
    char *t, *t2;
    int i = 0;
    int a = 0, sl = 0;

#ifdef AKII_DEBUG
    dbgenter(dbg_regqszavak, sz);
#endif
    if (!qszavak) {
#ifdef AKII_DEBUG
	dbgleave(dbg_regqszavak);
#endif
	return (-1);
    }
    torol(qszavak);
    strcpy(s, sz);
    t = malloc(strlen(s) + 2);
    do {
	p = s;
	while (isspace(*p))
	    p++;
	if (!(*p))
	    break;
	s = p;
	t2 = t;
	while ((!isspace(*p)) || a) {
	    if (!*p)
		break;
	    if ((*p == '"') && (!sl)) {
		a ^= 1;
		p++;
		continue;
	    }
	    if (*p == '\\') {
		sl ^= 1;
		if (!sl) {
		    *t2++ = '\\';
/**t2++='\\';*/
		}
		p++;
		continue;
	    }
	    if (sl) {
		sl = 0;
		if (*p != '"')
		    *t2++ = '\\';
	    }
	    *t2++ = *p;
	    p++;
	}
	if (sl) {
	    sl = 0;
	    *t2++ = '\\';
	}
	s = p;
	*t2 = 0;
	tedd(t, strlen(t), qszavak, i);
	i++;
    } while (*s);
    free(t);
#ifdef AKII_DEBUG
    dbgleave(dbg_regqszavak);
#endif
    return (i);
}
int regprobal(char *s, char *reg, REXELEM *hx)
{
    int i, helyszam = 0;
    char *p = reg, *q, *x;
    char tmp[256];
    char c;
    int before = 0, setbit, plus, szamol, kisnagy = 0;
    int maxrep, minrep;

#ifdef AKII_DEBUG
    dbgenter(dbg_regprobal, reg);
#endif
    if (hx)
	torol(hx);
    if (!(*s)) {
	if (*reg) {
	    if (strcmp(reg, "*") == 0 || strcmp(reg, "%") == 0) {
		if (hx)
		    tedd(s, 0, hx, helyszam);
		helyszam++;
#ifdef AKII_DEBUG
		dbgleave(dbg_regprobal);
#endif
		return (helyszam);
	    }
#ifdef AKII_DEBUG
	    dbgleave(dbg_regprobal);
#endif
	    return (-1);
	}
	else {
#ifdef AKII_DEBUG
	    dbgleave(dbg_regprobal);
#endif
	    return (0);
	}
    }
    while (1) {
	q = rstrchr(p, &c, &i);
	if ((!q) && (!*p)) {
	    if (!*s) {
#ifdef AKII_DEBUG
		dbgleave(dbg_regprobal);
#endif
		return (helyszam);
	    }
	    else {
		if (hx)
		    torol(hx);
#ifdef AKII_DEBUG
		dbgleave(dbg_regprobal);
#endif
		return (-1);
	    }
	}
	if (!q) {
	    if (rstrncmp(s, p, i + 1, kisnagy) == 0) {
#ifdef AKII_DEBUG
		dbgleave(dbg_regprobal);
#endif
		return (helyszam);
	    }
	    else {
		if (hx)
		    torol(hx); {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (-1);
		    }
	    }
	}
	if (p < q) {
	    if (rstrncmp(s, p, i, kisnagy)) {
		if (hx)
		    torol(hx); {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (-1);
		    }
	    }
	    s += i;
	    p = q;
	}
	else if ((c == '*') || (c == '%')) {
	    if (*++p == 0) {
		if (c == '*') {
		    if (hx)
			tedd(s, strlen(s), hx, helyszam);
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (++helyszam);
		}
		q = s;
		while (1) {
		    if (!*q) {
			if (hx)
			    tedd(s, strlen(s), hx, helyszam);
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (++helyszam);
		    }
		    if (*q++ < 33) {
			if (hx)
			    torol(hx); {
#ifdef AKII_DEBUG
			    dbgleave(dbg_regprobal);
#endif
			    return (-1);
			    }
		    }
		}
	    }
	    i = 0;
	    while (*p != 0) {
		if ((*p == '*') || (*p == '%') || (*p == '?') || (*p == '.') || (*p == '[') || (*p == '~') || (*p == '`'))
		    break;
		tmp[i++] = *p++;
		if (*(p - 1) == '\\')
		    tmp[i - 1] = *p++;
	    }
	    tmp[i] = 0;
	    if (i) {
		x = rstrstr(s, tmp, kisnagy);
		if (!x) {
		    if (hx)
			torol(hx); {
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (-1);
			}
		}
	    }
	    else {
		if (*p != '[') {
		    if (hx)
			torol(hx); {
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (-1);
			}
		}
		before = 1;
		continue;
	    }
	    if (hx)
		tedd(s, x - s, hx, helyszam);
	    helyszam++;
	    if (c == '%')
		while (s != x)
		    if (*s++ < 33) {
			if (hx)
			    torol(hx); {
#ifdef AKII_DEBUG
			    dbgleave(dbg_regprobal);
#endif
			    return (-1);
			    }
		    }
	    s = x + strlen(tmp);
	}
	else if ((c == '?') || (c == '.')) {
	    if (*s == 0) {
		if (hx)
		    torol(hx); {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (-1);
		    }
	    }
	    if ((c == '.') && (*s < 33)) {
		if (hx)
		    torol(hx); {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (-1);
		    }
	    }
	    if (*(p + 1) == '+') {
		if (isdigit(*(p + 2))) {
		    minrep = strtol(p + 2, &q, 10);
		    if (minrep < 1)
			minrep = 1;
		    if (*q == '-') {
			maxrep = strtol(q + 1, &q, 10);
		    }
		    else
			maxrep = minrep;
		    if (maxrep < minrep)
			maxrep = -1;
		    p = q - 1;
		}
		else {
		    minrep = maxrep = 0;
		    p++;
		}
		szamol = 1;
		c = *s;
		q = s + 1;
		while (*q++ == c) {
		    szamol++;
		    if (szamol == maxrep)
			break;
		}
		if ((minrep) && (szamol < minrep)) {
		    if (hx)
			torol(hx); {
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (-1);
			}
		}
	    }
	    else
		szamol = 1;
	    if (hx)
		tedd(s, szamol, hx, helyszam);
	    helyszam++;
	    if (!*(s += szamol)) {
		if (!*(p + 1)) {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (helyszam);
		}
		/*else {if (hx) torol(hx); return(-1);} */
	    }
	    p++;
	}
	else if (c == '[') {
	    p++;
	    if (*p == '^') {
		setbit = 0;
		p++;
	    }
	    else
		setbit = 1;
	    memset(tmp, setbit ^ 1, 256);
	    while (*p != ']') {
		if (*p == '\\')
		    p++;
		if (((*p == '`') || (*p == '~')) && (*(p - 1) != '\\')) {
		    p++;
		    kisnagy = (*(p - 1) == '`') ? 1 : 0;
		    continue;
		}
		if (!*p) {
		    if (hx)
			torol(hx); {
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (-1);
			}
		}
		if ((*p == '-') && (*(p - 1) != '[') && (*(p + 1) != ']') && (*(p + 1) != 0) && (*(p - 1) != '\\')) {
		    for (i = *(p - 1); i <= *(p + 1); i++) {
			tmp[i] = setbit;
			if (!kisnagy) {
			    tmp[toupper(i)] = setbit;
			    tmp[tolower(i)] = setbit;
			}
		    }
		    p += 2;
		}
		else {
		    if (*(p + 1) != '-') {
			tmp[*p] = setbit;
			if (!kisnagy) {
			    tmp[toupper(*p)] = setbit;
			    tmp[tolower(i)] = setbit;
			}
		    }
		    p++;
		}
	    }
	    if (*(p + 1) == '+') {
		p++;
		szamol = 0;
		plus = 1;
		if (isdigit(*(p + 1))) {
		    minrep = strtol(p + 1, &q, 10);
		    if (minrep < 0)
			minrep = 0;
		    if (*q == '-') {
			maxrep = strtol(q + 1, &q, 10);
		    }
		    else
			maxrep = minrep;
		    p = q - 1;
		    if (maxrep < minrep)
			maxrep = minrep;
		}
		else
		    minrep = -1;
	    }
	    else {
		plus = 0;
		szamol = 1;
	    }
	    if (!before) {
		if (!plus) {
		    if (!tmp[*s]) {
			if (hx)
			    torol(hx); {
#ifdef AKII_DEBUG
			    dbgleave(dbg_regprobal);
#endif
			    return (-1);
			    }
		    }
		}
		else {
		    q = s;
		    while (tmp[*q++]) {
			szamol++;
			if ((minrep >= 0) && (maxrep >= 0))
			    if (szamol >= maxrep)
				break;
		    }
		    if ((minrep >= 0) && (szamol < minrep)) {
			if (hx)
			    torol(hx); {
#ifdef AKII_DEBUG
			    dbgleave(dbg_regprobal);
#endif
			    return (-1);
			    }
		    }
		}
	    }
	    else {
		q = s;
		while (*q) {
		    if (tmp[*q]) {
			if (hx)
			    tedd(s, q - s, hx, helyszam);
			helyszam++;
			s = q;
			break;
		    }
		    q++;
		}
		if (!*q) {
		    if (hx)
			torol(hx); {
#ifdef AKII_DEBUG
			dbgleave(dbg_regprobal);
#endif
			return (-1);
			}
		}
		before = 0;
		if (plus) {
		    q = s;
		    while (tmp[*q++]) {
			szamol++;
			if ((maxrep >= 0) && (minrep >= 0))
			    if (szamol >= maxrep)
				break;
		    }
		    if ((maxrep >= 0) && (szamol < maxrep)) {
			if (hx)
			    torol(hx); {
#ifdef AKII_DEBUG
			    dbgleave(dbg_regprobal);
#endif
			    return (-1);
			    }
		    }
		}
	    }
	    if (hx)
		tedd(s, szamol, hx, helyszam);
	    helyszam++;
	    if (!*(s += szamol)) {
		if (!*(p + 1)) {
#ifdef AKII_DEBUG
		    dbgleave(dbg_regprobal);
#endif
		    return (helyszam);
		}
		/*else {if (hx) torol(hx); return(-1);} */
	    }
	    p++;
	}
	else if (c == '`') {
	    kisnagy = 1;
	    p++;
	}
	else if (c == '~') {
	    kisnagy = 0;
	    p++;
	}
    }
    if (hx)
	torol(hx);
#ifdef AKII_DEBUG
    dbgleave(dbg_regprobal);
#endif
    return (-1);
}
