/*
 *  TERMS, {akii-root}/src/terms.c
 *
 * ==========================================================================
 *
 *                      Arena Kliens II. Releases
 *
 * ==========================================================================
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A legkiralyabb szovegszerkeszto: VIM - Vi IMproved 5  <http://www.vim.org>
 */

int LINES = 24, COLS = 80;
int linuxbill = 0;

#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#ifdef _AIX
#include <sys/select.h>
#endif
#include "terms.h"

static struct termios T, eredeti;
static struct timeval tv;
static fd_set fs;

void setownterm(void)
{
    char *s;

    tcgetattr(0, &eredeti);
    tcgetattr(0, &T);
    T.c_lflag &= ~(ECHO + ICANON + ISIG);
    T.c_iflag &= ~(IXON);
    T.c_cc[VTIME] = 0;
    T.c_cc[VMIN] = 0;
    tcsetattr(0, TCSANOW, &T);
    s = getenv("TERM");
    if (!s)
	return;
    if ((strcmp(s, "linux") == 0) || (strcmp(s, "LX") == 0))
	linuxbill = 1;
    return;
}
void setoriginalterm(void)
{
    tcsetattr(0, TCSANOW, &eredeti);
    return;
}
void getwinsize(void)
{
    struct winsize W;

    if (ioctl(0, TIOCGWINSZ, &W) == 0) {
	if (!(LINES = W.ws_row))
	    LINES = 24;
	if (!(COLS = W.ws_col))
	    COLS = 80;
    }
    else {
	LINES = 24;
	COLS = 80;
    }
}

int getch(void)
{
    unsigned char i, j;
    int r;

    r = read(0, &i, 1);
    if (r <= 0)
	return (r);
    if (i == 127)
	i = 8;
    if (i != 27)
	return (i);

    tv.tv_sec = 0;
    tv.tv_usec = 500000;
    FD_ZERO(&fs);
    FD_SET(0, &fs);
    if (select(1, &fs, NULL, NULL, &tv) <= 0)
	return (i);
    r = read(0, &i, 1);
    if (r <= 0)
	return (r);
    if ((i != '[') && (i != 'O')) {
	if (((i >= 'a') && (i <= 'z')) || ((i >= '0') && (i <= '9')))
	    return (KEY_ALT(i));
	return (i);
    }

    tv.tv_sec = 0;
    tv.tv_usec = 10000;
    FD_ZERO(&fs);
    FD_SET(0, &fs);
    if (select(1, &fs, NULL, NULL, &tv) <= 0)
	return (i);
    r = read(0, &i, 1);
    if (r <= 0)
	return (r);
    switch (i) {
	case 'A':
	    return (KEY_UP);
	case 'B':
	    return (KEY_DOWN);
	case 'C':
	    return (KEY_RIGHT);
	case 'D':
	    return (KEY_LEFT);
    }
    if (!((i >= '0') && (i <= '8')))
	return (i);

    tv.tv_sec = 0;
    tv.tv_usec = 500000;
    FD_ZERO(&fs);
    FD_SET(0, &fs);
    if (select(1, &fs, NULL, NULL, &tv) <= 0)
	return (i);
    r = read(0, &j, 1);
    if (r <= 0)
	return (r);
    if (j != '~') {
	j = (i - 48) * 100 + (j - 48) * 10;
	tv.tv_sec = 0;
	tv.tv_usec = 10000;
	FD_ZERO(&fs);
	FD_SET(0, &fs);
	if (select(1, &fs, NULL, NULL, &tv) <= 0)
	    return (j);
	r = read(0, &i, 1);
	if (r <= 0)
	    return (r);
	j += i - 48;
	tv.tv_sec = 0;
	tv.tv_usec = 10000;
	FD_ZERO(&fs);
	FD_SET(0, &fs);
	if (select(1, &fs, NULL, NULL, &tv) <= 0)
	    return (j);
	r = read(0, &i, 1);
	if (r <= 0)
	    return (r);
	if ((i != 'q') && (i != 'z'))
	    return (j);
	switch (j) {
	    case 214:
	    case 90:
		return (KEY_HOME);
	    case 247:
	    case 81:
		return (KEY_INS);
	    case 77:
		return (KEY_DEL);
	    case 220:
	    case 88:
		return (KEY_END);
	    case 216:
	    case 83:
		return (KEY_PGUP);
	    case 222:
	    case 106:
		return (KEY_PGDN);
	    case 139:
		return (KEY_INS);
	}
	return (j);
    }
    switch (i) {
	case '1':
	    if (linuxbill)
		return (KEY_HOME);
	    else
		return (KEY_INS);
	
	case '2':
	    if (linuxbill)
		return (KEY_INS);
	    else
		return (KEY_HOME);
	
	case '3':
	    if (linuxbill)
		return (KEY_DEL);
	    else
		return (KEY_PGUP);
	
	case '4':
	    if (linuxbill)
		return (KEY_END);
	    else
		return (KEY_DEL);
	
	case '5':
	    if (linuxbill)
		return (KEY_PGUP);
	    else
		return (KEY_END);
	
	case '6':
	    return (KEY_PGDN);

	case '7':
	    return (KEY_INS);
	
	case '8':
	    return (KEY_DEL);
    }
    return (i);
}
